#ifndef LAUNCHELF_H
#define LAUNCHELF_H

#include <stdio.h>
#include <tamtypes.h>
#include <sifcmd.h>
#include <kernel.h>
#include <sifrpc.h>
#include <loadfile.h>
#include <string.h>
#include <malloc.h>
#include <libhdd.h>
#include <libmc.h>
#include <libpad.h>
#include <fileio.h>
#include <sys/stat.h>
#include <iopheap.h>
#include <errno.h>
#include <fileXio_rpc.h>
#include <iopcontrol.h>
#include <stdarg.h>
#include <sbv_patches.h>
#include <sys/fcntl.h>
//#include <debug.h>
#include <ito.h>
#include <cdvd_rpc.h>
#include <smod.h>
#include "cd.h"
#include "mass_rpc.h"

#include "language.h"
#include "cnf.h"

// o[W
#define LBF_VER "LbF v0.52"

// XL[g
#define SCANRATE (ITO_VMODE_AUTO==ITO_VMODE_NTSC ? 60:50)

#define UNK_BOOT 0
#define CD_BOOT 1
#define MC_BOOT 2
#define HOST_BOOT 3
#define PFS_BOOT 4
#define VFS_BOOT 5
#define HDD_BOOT 6
#define MASS_BOOT 7

enum
{
	SCREEN_WIDTH = 640,
//	SCREEN_HEIGHT = 448,
//	SCREEN_MARGIN = 14,
//	FONT_WIDTH = 10,
//	FONT_HEIGHT = 20,
	LINE_THICKNESS = 2,
	
	MAX_NAME = 256,
	MAX_PATH = 1025,
	MAX_ENTRY = 2048,
//	MAX_ROWS = 16,
	MAX_PARTITIONS=100
};

typedef struct
{
	char dirElf[13][MAX_PATH];
	int timeout;
	int filename;
	uint64 color[8];
	int screen_x;
	int screen_y;
	int discControl;
	int flickerControl;
	int interlace;
	int ffmode;
	int fileicon;
	int discPs2saveCheck;
	int discELFCheck;
	char Exportdir[MAX_PATH];
	int language;
	char AsciiFont[MAX_PATH];
	char KanjiFont[MAX_PATH];
	int CharMargin;
	int LineMargin;
	int FontBold;
	int AsciiMarginTop;
	int AsciiMarginLeft;
	int KanjiMarginTop;
	int KanjiMarginLeft;
} SETTING;

/* main.c */
extern char LaunchElfDir[MAX_PATH], LastDir[MAX_NAME];
extern int boot;
void loadCdModules(void);
void loadUsbModules(void);
void loadHddModules(void);

/* elf.c */
int checkELFheader(const char *filename);
void RunLoaderElf(char *filename, char *);

/* draw.c */
//setupito
enum{
	ITO_INIT_ENABLE,
	ITO_INIT_DISABLE
};

//SetFontMargin GetFontMargin
enum
{
	CHAR_MARGIN = 0,
	LINE_MARGIN,
	ASCII_FONT_MARGIN_TOP,
	ASCII_FONT_MARGIN_LEFT,
	KANJI_FONT_MARGIN_TOP,
	KANJI_FONT_MARGIN_LEFT
};

//GetCurrentPos
enum
{
	CURRENTPOS_X = 0,
	CURRENTPOS_Y
};

//GetFontSize
enum
{
	ASCII_FONT_WIDTH = 0,
	ASCII_FONT_HEIGHT,
	KANJI_FONT_WIDTH,
	KANJI_FONT_HEIGHT
};

extern itoGsEnv screen_env;
extern int SCREEN_HEIGHT;
extern int SCREEN_MARGIN;
extern int FONT_WIDTH;
extern int FONT_HEIGHT;
extern int MAX_ROWS;
void drawDark(void);
void drawDialogTmp(int x1, int y1, int x2, int y2, uint64 color1, uint64 color2);
void setScrTmp(const char *msg0, const char *msg1);
void drawMsg(const char *msg);
void setupito(int flag);
void clrScr(uint64 color);
void drawScr(void);
void SetHeight(void);
void drawFrame(int x1, int y1, int x2, int y2, uint64 color);
int InitFontAscii(const char *path);
int InitFontKnaji(const char *path);
void FreeFontAscii(void);
void FreeFontKnaji(void);
int SetFontMargin(int type, int Margin);
int GetFontMargin(int type);
int GetFontSize(int type);
void SetFontBold(int flag);
int GetFontBold(void);
int checkFONTX2header(const char *path);
void drawChar(unsigned char c, int x, int y, uint64 colour);
int printXY(const unsigned char *s, int x, int y, uint64 colour, int draw);
int SetCurrentPos(int x, int y);
int GetCurrentPos(int type);
int printXY2(const unsigned char *s, uint64 color, int draw);

/* pad.c */
extern u32 new_pad;
int setupPad(void);
int readpad(void);
void waitPadReady(int port, int slot);

/* config.c */
extern SETTING *setting;
void loadConfig(char *);
void config(char *);

/* filer.c */
enum	//getFilePath
{
	ANY_FILE = 0,
	ELF_FILE,
	DIR,
	FNT_FILE
};
void MessageDialog(const char *message);
int newdir(const char *path, const char *name);
int keyboard(char *out, int max);
void getFilePath(char *out, const int cnfmode);

/* language.c */
extern LANGUAGE *lang;
void InitLanguage(void);
void FreeLanguage(void);
void SetLanguage(const int langID);

#endif
